`timescale 1ns / 10ps

module intel_10c_frontend(gtx_refclk_p, gtx_rxp, gtx_txp, pipe_rx_clk,
  pipe_clk, raw_tx, raw_rx, pll_locked, frontend_rst, refclk_locked,
  mgt_powerdown, ltssm_lowpower, rx_dontmess, rx_phy_ready, rx_resync,
  rx_detect_revpolarity, rx_detect_revpolarity_clear, rx_align_enable,
  tx_elecidle, receiver_detect, receiver_present, receiver_present_valid,
  version_10c_1_0);

  input  gtx_refclk_p;
  input  gtx_rxp;
  input [39:0] raw_tx;
  input  mgt_powerdown;
  input  ltssm_lowpower;
  input  rx_dontmess;
  input  rx_resync;
  input  rx_detect_revpolarity;
  input  rx_align_enable;
  input  tx_elecidle;
  input  receiver_detect;
  output  gtx_txp;
  output  pipe_rx_clk;
  output  pipe_clk;
  output [39:0] raw_rx;
  output  pll_locked;
  output  frontend_rst;
  output  refclk_locked;
  output  rx_phy_ready;
  output  rx_detect_revpolarity_clear;
  output  receiver_present;
  output  receiver_present_valid;
  output  version_10c_1_0;
  reg  pll_locked;
  reg  rx_phy_ready;
  reg  rx_detect_revpolarity_clear;
  reg [3:0] mgt_powerdown_d;
  reg  rxpolarity;
  reg [6:0] polarity_debounce_cnt;
  reg [4:0] wakeup_counter;
  reg  reset_sync;
  reg [127:0] tx_parallel_data;
  reg  rx_digitalreset;
  reg  rx_digitalreset_d;
  reg  rx_digitalreset_d2;
  reg  rx_set_locktodata;
  reg  lockedtoref_d;
  reg  lockedtoref_d2;
  reg  rx_resync_latch;
  reg  ltssm_lowpower_d;
  reg  ltssm_lowpower_d2;
  reg  rx_resync_d;
  reg  rx_resync_d2;
  reg  rx_resync_unlatch;
  reg  rx_resync_unlatch_d;
  reg  rx_resync_unlatch_d2;
  reg  rx_dontmess_d;
  reg  rx_dontmess_d2;
  reg  rx_align_enable_d;
  reg  rx_align_enable_d2;
  reg  xcvr_pll_locked_d;
  reg [1:0] rx_state;
  reg [10:0] ltr_ltd_cnt;
  reg [17:0] datalock_cnt;
  wire  mgmt_clk;
  wire  xcvr_pll_locked;
  wire  rstctl_rx_digitalreset;
  wire [127:0] rx_parallel_data;
  wire  rx_is_lockedtoref;

   localparam
     st_start = 0,
     st_wait_reflock = 1,
     st_datalock_wait = 2,
     st_datalock = 3;

   initial wakeup_counter = 0;

   assign version_10c_1_0 = 1;

   assign frontend_rst = !pll_locked;

   assign refclk_locked = 1;

   assign mgmt_clk = gtx_refclk_p;

   assign receiver_present = 0;
   assign receiver_present_valid = 0;

   always @(*)
     begin
	tx_parallel_data <= 0; // Default to zero

	tx_parallel_data[42:33] <= raw_tx[39:30];
	tx_parallel_data[31:22] <= raw_tx[29:20];
	tx_parallel_data[20:11] <= raw_tx[19:10];
	tx_parallel_data[9:0] <= raw_tx[9:0];
     end

   assign raw_rx = { rx_parallel_data[57:48],
		     rx_parallel_data[41:32],
		     rx_parallel_data[25:16],
		     rx_parallel_data[9:0] };

   always @(posedge mgmt_clk)
     begin
	if (!(&wakeup_counter))
	  wakeup_counter <= wakeup_counter + 1;

	// Reset transceiver (also) when waking up from low-power mode.
	if (mgt_powerdown_d[3] && !mgt_powerdown_d[2])
	  wakeup_counter <= 0;

	reset_sync <= !(&wakeup_counter);

	mgt_powerdown_d <= { mgt_powerdown_d, mgt_powerdown };

	lockedtoref_d <= rx_is_lockedtoref;
	lockedtoref_d2 <= lockedtoref_d;

	rx_resync_d <= rx_resync_latch;
	rx_resync_d2 <= rx_resync_d;

	ltssm_lowpower_d <= ltssm_lowpower;
	ltssm_lowpower_d2 <= ltssm_lowpower_d;

	rx_resync_unlatch <= (rx_state == st_start);

	rx_dontmess_d <= rx_dontmess;
	rx_dontmess_d2 <= rx_dontmess_d;

	rx_align_enable_d <= rx_align_enable;
	rx_align_enable_d2 <= rx_align_enable_d;

	rx_digitalreset <= rstctl_rx_digitalreset ||
			   ((rx_state != st_datalock) && !ltssm_lowpower_d2);
	rx_set_locktodata <= ((rx_state == st_datalock_wait) ||
			      (rx_state == st_datalock));

	// Because of the -5000 ppm SSC invovled, the transceiver's data lock
	// detector is useless. To work around this, loop on lock-to-reference
	// and lock-to-data back and forth, giving the lock-to-data part
	// 1 ms to lock on the data before releasing the digital reset.
	// rx_align_enable is used to indicate proper data arriving
	// (i.e. a data lock), but if rx_dontmess is asserted, loop anyway
	// (this means that the LTSSM isn't expecting any data anyhow).

	// In each loop, proper data is let through (if present) for 200 us,
	// which is enough to flip polarity if necessary and detect ordered
	// sets, so both rx_align_enable and rx_dontmess can be
	// deasserted in a timely manner.

	case (rx_state)
	  st_start:
	    begin
	       ltr_ltd_cnt <= 0;
	       datalock_cnt <= 0;

	       rx_state <= st_wait_reflock;
	    end

	  st_wait_reflock:
	    // Assuming mgmt_clk @ 125 MHz, tLTR_LTD_manual = 15us => 1875
	    if (!lockedtoref_d2)
	      ltr_ltd_cnt <= 0;
	    else if (ltr_ltd_cnt == 1875)
	      rx_state <= st_datalock_wait;
	    else
	      ltr_ltd_cnt <= ltr_ltd_cnt + 1;

	  st_datalock_wait:
	    if (datalock_cnt == 125000) // 1 ms @ 125 MHz
	      rx_state <= st_datalock;
	    else
	      datalock_cnt <= datalock_cnt + 1;

	  st_datalock:
	    if (datalock_cnt != 150000) // 200 us into this state @ 125 MHz
	      datalock_cnt <= datalock_cnt + 1;
	    else
	      begin
		 if (!rx_dontmess_d2)
		   rx_state <= st_start; // Periodic self-reset

		 // A need for comma alignment any time after 200 us is
		 // interpreted as no lock or being torn from lock. Try again.
		 if (rx_align_enable_d2)
		   rx_state <= st_start;
	      end
	endcase

	// Reset overrides state assignments above
	if (rstctl_rx_digitalreset || rx_resync_d2)
	  rx_state <= st_start;
     end

   always @(posedge pipe_clk)
     begin
	xcvr_pll_locked_d <= xcvr_pll_locked;
	pll_locked <= xcvr_pll_locked_d;

	// Handle rx polarity inversion

	rx_detect_revpolarity_clear <= (polarity_debounce_cnt != 0);

	if (polarity_debounce_cnt != 0)
	  polarity_debounce_cnt <= polarity_debounce_cnt - 1;
	else if (rx_detect_revpolarity)
	  begin
	     polarity_debounce_cnt <= polarity_debounce_cnt - 1; // Wrap to max
	     rxpolarity <= !rxpolarity;
	  end

	rx_resync_unlatch_d <= rx_resync_unlatch;
	rx_resync_unlatch_d2 <= rx_resync_unlatch_d;

	rx_digitalreset_d <= rx_digitalreset;
	rx_digitalreset_d2 <= rx_digitalreset_d;

	if (rx_resync)
	  rx_resync_latch <= 1;
	else if (rx_resync_unlatch_d2)
	  rx_resync_latch <= 0;

	// Even if rx_digitalreset_d2 is asserted one clock after
	// rx_resync_unlatch_d2 is asserted due to differences in
	// clock domain crossing's sampling, rx_resync_latch adds an extra
	// delay, so rx_phy_ready remains low on this handover.

	rx_phy_ready <= !rx_digitalreset_d2 && !rx_resync_latch;
     end

   xcvr xcvr_inst
   (
    .tx_pll_refclk_clk(gtx_refclk_p),
    .rx_pll_refclk_clk(gtx_refclk_p),

    .tx_serial_data_tx_serial_data(gtx_txp),
    .rx_serial_data_rx_serial_data(gtx_rxp),

    .tx_parallel_data_tx_parallel_data(tx_parallel_data),
    .rx_parallel_data_rx_parallel_data(rx_parallel_data),

    .tx_clkout_clk(pipe_clk),
    .tx_coreclkin_clk(pipe_clk),
    .rx_clkout_clk(pipe_rx_clk),
    .rx_coreclkin_clk(pipe_rx_clk),

    .rx_set_locktodata_rx_set_locktodata(rx_set_locktodata),
    .rx_set_locktoref_rx_set_locktoref(1'b1), // Default; overridden by rx_set_locktodata
    .rstctl_rx_is_lockedtodata_rx_is_lockedtodata(1'b1), // Ignored by reset controller
    .rx_is_lockedtodata_rx_is_lockedtodata(), // Unreliable with 0.5% SSC, hence unused
    .rx_is_lockedtoref_rx_is_lockedtoref(rx_is_lockedtoref),

    .tx_polinv_tx_polinv(1'b0), // No Tx polarity inversion
    .rx_polinv_rx_polinv(rxpolarity),

    // Management ports
    .mgmt_clk_clk(mgmt_clk),
    .reset_reset(reset_sync),

    .pll_locked_pll_locked(xcvr_pll_locked),
    .rstctl_pll_locked_pll_locked(xcvr_pll_locked),
    .pll_select_pll_select(1'b0), // Single PLL
    .rstctl_rx_digitalreset_rx_digitalreset(rstctl_rx_digitalreset),
    .rx_digitalreset_rx_digitalreset(rx_digitalreset)
    );

endmodule
